/*********************************************************************
 * File:   install.c  (part of TTYTALK)
 * Author: Victor Volkman
 * Date:   June, 1988
 ********************************************************************/
#include "tty.h"

#define SET_INT 0x25
#define GET_INT 0x35
#define VECSTART 8

extern struct serial_def serial;
extern int iobase;
extern int q_errs, r_errs;

int ser_interrupts[] = {PRIMARY_INT, SECONDARY_INT};
int ser_ports[] = {PRIMARY_IO, SECONDARY_IO};
int ssave, osave;

void install(vector,new_seg,new_offs,old_seg,old_offs)
int vector;
int new_seg;
int new_offs;
int *old_seg;
int *old_offs;
{
#if defined MSC
    union REGS r;
    struct SREGS s;
#else
    union REGSS r;      /* Complete 8086 register set    */
#endif

    r.h.ah = GET_INT;       /* Get old interrupt vector from DOS */
    r.h.al = vector;    /* and return result in ES:BX    */

#if defined MSC
    intdosx(&r, &r, &s);
    *old_seg = s.es;        /* Save the old interrupt vector so  */
    *old_offs = r.x.bx;     /* we can restore it later on    */

#else
    intdoss(&r,&r);
    *old_seg = r.x.es;      /* Save the old interrupt vector so  */
    *old_offs = r.x.bx;     /* we can restore it later on    */
#endif

    r.h.ah = SET_INT;       /* Set new interrupt vector via DOS  */
    r.h.al = vector;    /* with address in DS:DX     */
    r.x.dx = new_offs;

#if defined MSC
    s.ds = new_seg;
    intdosx(&r, &r, &s);
#else
    r.x.ds = new_seg;
    intdoss(&r,&r);
#endif
}

#if defined MSC
far interrupt enqueue(void);
#else
void int_serv(void);
#endif

void start_rcv()       /* begin communications with ISR */
{
    r_errs = 0;         /* clear count of receiver errors    */
    q_errs = 0;         /* clear count of queue overflows    */

    iobase = ser_ports[serial.adapter];    /* I/O address of 8250 */
    comm_setup();       /* Setup communications via INT 14h  */
#if defined MSC
    enable_isr(enqueue);
#else
    enable_isr(int_serv);
#endif
}


void stop_rcv()        /* end communicatons with ISR    */
{
    disable_isr();
}

/** Never run disable_isr unless enable_isr already called */
static int isr_enabled = 0;

#if defined MSC
void enable_isr(pfun)
void (far *pfun) ();
#else
void enable_isr(pfun)      /* enable ISR function for 8250  */
void (*pfun) ();
#endif
{
    int int_no;         /* hardware INT we are installing    */
    int mcr;        /* 8250 modem control register   */
    int iseg;           /* segment of pfun() ISR     */
    int ioff;           /* offset of pfun() ISR      */

    int_no = ser_interrupts[serial.adapter];
#if defined MSC
    iseg = (int)( ((long)pfun) >> 16 );   /* small model segment of function */
    ioff = (int)( (long) pfun );
#else
    iseg = get_cs();      /* small model segment of function */
    ioff = FP_OFF((long)pfun);    /* small model offset of function  */
#endif
    install(int_no+VECSTART,iseg,ioff,&ssave,&osave);
    set_8259(int_no);         /* Enable hardware interrupt       */
    outp(iobase+INT_ENABLE,ENABLE_RCV_INT);  /* Enable 8250 RxRDY    */

    mcr = inp(iobase+MODEM_CTRL) | GPO2;     /* Enable 8250 Master   */
    outp(iobase+MODEM_CTRL,mcr);         /* interrupt control    */
    /** never run disable_isr unless enable_isr already called */
    isr_enabled = 1;
}



void disable_isr()
{
    int mcr, int_no, temp;

    /** never run disable_isr unless enable_isr already called */
    if (!isr_enabled)
        return;

    int_no = ser_interrupts[serial.adapter];
    clear_8259(int_no);

    mcr = inp(iobase+MODEM_CTRL) & (~GPO2);   /* Disable 8250 Master */
    outp(iobase+MODEM_CTRL,mcr);          /* interrupt control   */

    outp(iobase+INT_ENABLE,0);            /* Disable 8250 RxRDY  */
    clear_8259(int_no);
    install(int_no+VECSTART,ssave,osave,&temp,&temp);
}

#define MASK_PORT 0x21
#define EOI_PORT  0x20
#define EOI_CMD   0x20

void set_8259(int_no)      /* set a bit in 8259 IMR */
int int_no;
{
    unsigned imr;     /* interrupt mask register */

    imr = inp(MASK_PORT);
    imr = imr & ~ (1 << int_no);
    outp(MASK_PORT,imr);
}


void clear_8259(int_no)    /* clear a bit in 8259 IMR */
int int_no;
{
    unsigned imr;     /* interrupt mask register */

    imr = inp(MASK_PORT);
    imr = imr | (1 << int_no);
    outp(MASK_PORT,imr);
}
